// c. 2015 Technologic Systems
// Written by Michael D. Peters
// This is example code intended for no practical purpose.
// This code is for educational intent only.
//
// Use of this code indcates acceptance of sole
// responsibility for any outcome intended or 
// otherwise.

// Implementation notes:
// TS-ADC16 
// FIFO Depth 512 samples.  More channels per sample = faster FIFO fill.
// State machine will stop when FIFO overflows.  
// FIFO empties only in 3 conditions:
//   1.  Exttrig started state machine.
//   2.  SBC read out all data from FIFO.
//   3.  System power loss.
//
// Register 0x0  ID
// bits 15:12 -- Jumper status (jp4, jp3, jp2, jp1)
// bits 11:8  -- PLD Revision (Expect 5+)
// bits 7:0   -- Peripheral ID:  Expect 0x3E
//
// Register 0x2  ADCCFG 
// bit 9 exttrig  (start adc on external signal)
//                NOTE:  This trigger can only *start*
//                       the state machine, it cannot 
//                       *stop* the state machine.
// bit 8 signle/differential 1=single 0=diff
// bit 7:6 input range 01 = 0-5v (and other values see the TS-ADC16 wiki)
// bit 5 single/differential -- must be set same as bit 8
// bit 4:1 numchan -- number of channel pairs 0000 = 1 pair
// bit 0 start/stop - start or stop peripheral state machine
//
// Register 0x4  ADCDLY_MSB
// Upper 16 bits of ADCDLY
//
// Register 0x6  ADCDLY_LSB
// Lower 16 bits of ADCDLY
//
// Register 0x8  ADCSTAT
// bits 15:6 - Read -- FFCOUNT - Number of samples sitting in FIFO
//        - Write-- FFCOUNT - Sample count required to trigger IRQ
// bits 5:1  -- FFHEAD - ID of channel for sample at head of FIFO.
// bit  0    - INTEN - Enable(1)/disable(0) interrupt generation for FFCOUNT(Write)
//
// Register 0xA  ADCFIFO
// 16 bit ADC read data from channel FFHEAD.  Consumes sample, decrements FFCOUNT(Read).
//
// Register 0xE  DAC control register
// bits 15:14 -- DAC number to set (0, 1, 2, 3)
// bit  13    -- DAC Max value:  Set 0 for 2.5V, set 1 for 5V
// bit  12    -- Always write 1 here.
// bits 11:0  -- 12 bit DAC level
//  
// Jumpers:
// JP4 OFF means device IRQ 6.
// JP4 ON  means device IRQ 7.

//
// TS-7800 esotera
// PC104 setup:
/* 
 0x55555555 to address 0xE8000030
 0x55555555 to address 0xE8000034
 0x55555 to address 0xE8000038
 0x55555 to address 0xE800003C 
 */
// PC104 16 bit base address is 0xEF000000
// PC104 IRQ6 is TS-7800 IRQ69
// PC104 IRQ7 is TS-7800 IRQ70
// 
// This code requires "userspace IRQ".  Be sure your kernel has this function.
// The image here
//  ftp://ftp.embeddedarm.com/ts-arm-sbc/ts-7800-linux/binaries/ts-images/512mbsd-latest.dd.bz2
//  has an appropriately patched Linux 2.6.29 kernel.
//
// Compile on SBC thus:
// gcc <thisfile> -o <executable> -mcpu=arm9 -DTS7800
//
// Wiring the hardware:
//  TS-7800 with a TS-ADC16 on the PC104 header.
//  TS-ADC16 with JP3 and optionally JP4 (IRQ6 or IRQ7)
//  DAC-A to AD1:0  (screw terminals)
//  DAC-B to AD2:0  (screw terminals)
//  DAC-C to AD1:1  (screw terminals)
//  DAC-D to AD2:1  (screw terminals)
//  One jumper wire on AD1 header pin 19.  Connect this to pin 20 when you want to
//   start the state machine.  This is the external trigger input.

#include <stdio.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdint.h>
#include <assert.h>
#include <errno.h>
#include <sched.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <sys/time.h>

#define PC104BASE 0xEF000000

uint16_t lsb32(uint32_t);  // Returns the lower 16 bits of a 32 bit number.
uint16_t msb32(uint32_t);  // Returns the upper 16 bits of a 32 bit number.

// Assumes TS-7800 + TS-ADC16
// Sets TS-ADC16 to external trigger start, single-ended, IRQ enabled at 16 sample FIFO.
int main(int argc, char* argv[]){
  int a,b,x,y,i = 0;  // counter variables
  int pc104id = 0;
  unsigned int sample_count = 0;
  uint16_t irqnumber = 0;
  int irq_handle = 0;
  volatile uint16_t *pc104_base = 0;
  volatile uint32_t *membase = 0;
  uint16_t adc16_base  = 0x100; // expected ADC16 address offset.
  volatile uint16_t *adc16_io = 0;  // pointer to the ADC16 hardware.
  uint32_t adc_dly = 0;  // sample delay = 31.25ns * adc_dly.
  int mem,mem2 = 0;         // file designator used in open()
  struct sched_param sp = { .sched_priority=99 };  // Scheduler high priority
  char proc_irq[17]; // String container for "/proc/irq/nn/irq".
  fd_set fds;
  struct timeval tv;
  struct timespec ts;

  printf("Sched priority requested is %d.\n", sp.sched_priority);

  // Declare realtime priority to the OS.
  a = sched_setscheduler(0, SCHED_RR, &sp);
  if(a == -1) {
    fprintf(stderr, "WARNING:  ADC reader Process did not receive priority, \"%s\"\n", (strerror(errno)));
  }

  sched_rr_get_interval(0, &ts);
  printf("SCHED_RR Time quantum %d seconds, %d nseconds.\n", ts.tv_sec, ts.tv_nsec);

  // Open and set up device pointers.
  mem = open("/dev/mem", O_RDWR|O_SYNC);
  pc104_base = mmap(0, getpagesize(), PROT_READ | PROT_WRITE, 
                    MAP_SHARED, mem, PC104BASE);
  adc16_io = pc104_base + (adc16_base/2);

#ifdef TS7800
  // Make sure PC104 is turned on if we're using a TS-7800 SBC.
  mem2 = open("/dev/mem", O_RDWR|O_SYNC);
  membase = mmap(0, getpagesize(), PROT_READ | PROT_WRITE,
                 MAP_SHARED, mem2, 0xE8000000);
  membase[0x30/4] = 0x55555555;
  membase[0x34/4] = 0x55555555;
  membase[0x38/4] = 0x55555;
  membase[0x3C/4] = 0x55555;
#endif

  // Verify hardware presence.
  pc104id = adc16_io[0];
  assert((pc104id & 0x00FF) == 0x003E);  // if board ID is not 0x3E, quit loudly.

  // Set up IRQ handling.
  irqnumber = (pc104id>>15)?69:70;  // Discover which IRQ to watch for.  -- assuming TS-7800
  snprintf(proc_irq, sizeof(proc_irq), "/proc/irq/%d/irq", irqnumber); // Populate irq file path.
  irq_handle = open(proc_irq, O_RDONLY | O_SYNC | O_NONBLOCK, S_IRUSR); // Open proc irq, get file handle.
  if(irq_handle == -1){
    fprintf(stderr,"Unable to open IRQ %d:  %s\n", irqnumber, strerror(errno));
    assert(irq_handle == -1);  // No IRQ.  Quit.  Loudly.
  }
  FD_ZERO(&fds);
  
  // Set starting ADC setup values.
  adc_dly = 500000;  // 32000000/ADC_DLY = sample frequency in Hz.  Bigger numbers here mean slower FIFO fill-up.
  adc16_io[2/2] = 0x362; // Set up state machine.
  adc16_io[4/2] = msb32(adc_dly);
  adc16_io[6/2] = lsb32(adc_dly);
  adc16_io[8/2] = 0x801; // trigger IRQ on (n>>6) samples.

  // Turn on DAC #0, set it to 5VDC.
  adc16_io[0xE/2] = 0x3FFF;
  usleep(100000);  // Give a little time between these writes so the SBC can write to the TS-ADC16 which will write to the DAC chip.
  // Turn on DAC #1, set it to 3.75VDC.
  adc16_io[0xE/2] = 0x7BFD;
  usleep(500);     // Once it gets going, you can speed up a lot, but still give it some time.
  // Turn on DAC #2, set it to 2.5 VDC.
  adc16_io[0xE/2] = 0xB7FE;
  usleep(500);
  // Turn on DAC #3, set it to 1.25VDC.
  adc16_io[0xE/2] = 0xF3FF;
  usleep(500);
  
  // SBC State machine infinite loop.
  while(1){
    // Set select() time out to 1 second.
    tv.tv_sec = 1;
    tv.tv_usec = 0;

    // Wait for IRQ.
    FD_SET(irq_handle, &fds);  // put the IRQ file handle in the watched file set.
    y = select(FD_SETSIZE, &fds, NULL, NULL, &tv); // Check IRQ status
    if(y == 0){
      printf("IRQ timeout on ADC read.  FIFO Count at %d.\n", (adc16_io[8/2]>>6));  // This can be on purpose.  Check your settings if you didn't expect to see this.
      fflush(stdout);
    }
    else if(FD_ISSET(irq_handle, &fds)){
      // do irq triggered stuff.
      read(irq_handle, &y, sizeof(y)); // read out irq junk data.
      sample_count = adc16_io[8/2] >> 6;
      printf("IRQ Triggered.  FIFO count: %d.  Trigger %d usec.\n", sample_count, (1000000 - ((tv.tv_sec * 1000000) + (tv.tv_usec)))); // tv is set to time-not-slept by select().
      fflush(stdout);

      // Read out all desired FIFO data.  Ideally don't use printf because it's slow and you don't have a lot of time - that FIFO can fill up FAST if you set it up right.
      for(i = 0; i < sample_count; i++){
        printf("Channel=%d\nvalue=0x%4X\n---\n", (( adc16_io[0x8/2] & 0x001F) >> 1  ), adc16_io[0xA/2]); // FFHEAD@0x8 bits 5:1, FIFO@0xA bits 15:0.
        fflush(stdout);
      }
    }    
    usleep(100);  // This usleep is ABSOLUTELY NECESSARY to prevent an unwanted long pre-emption.
  }
  return 0;
}

// Returns the ls16b of a 32b number as a 16b number.
uint16_t lsb32(uint32_t number)
{
    uint16_t lsb = 0;

    lsb = (uint16_t)((uint32_t)(number & 0x0000FFFF));
    
    return lsb;
}

// Returns the ms16b of a 32b number as a 16b number.
uint16_t msb32(uint32_t number)
{
    uint16_t msb = 0;
    
    msb = (uint16_t)((uint32_t)(number & 0xFFFF0000) >> 16);
    
    return msb;
}
